import re
from datetime import date, datetime
from typing import Collection, Iterable, Optional, Tuple, cast

from bs4 import BeautifulSoup, Tag

from .entities import Event
from .event_parser import EventParser
from .fetching import BaseWebFetcher

DATE_RE = re.compile(
    r"(?P<month>\w{3,4})\s+(?P<start_day>\d{1,2})"
    r"(\s?[-–—]\s?(?P<end_month>\w{3,4})?\s*(?P<end_day>\d{1,2}))?"
)


def date_for_components(
    month: str, day: str, year: int, wrap_after: int
) -> Optional[date]:
    try:
        parsed = datetime.strptime(f"{month} {day} {year}", "%b %d %Y").date()
        if wrap_after < parsed.month:
            parsed = parsed.replace(year=parsed.year - 1)
        return parsed
    except ValueError:
        return None


def parse_html_dates(
    date_string: str, base_year: int, wrap_after: int = 13
) -> Collection[Tuple[date, date]]:
    dates = []

    for possible_date in DATE_RE.finditer(date_string):
        month = possible_date.group("month")[:3]
        start_day = possible_date.group("start_day")
        end_day = possible_date.group("end_day")
        start_date = date_for_components(month, start_day, base_year, wrap_after)
        if start_date is None:
            continue
        end_date = None

        if end_day:
            end_month = possible_date.group("end_month")
            if end_month:
                end_month = end_month[:3]
            else:
                end_month = month
            end_date = date_for_components(end_month, end_day, base_year, wrap_after)

        if end_date is None:
            end_date = start_date

        if start_date:
            dates.append((start_date, end_date))

    return dates


class HTMLFetcher(BaseWebFetcher):
    base_url = "https://ro.umich.edu/calendars"


class HTMLEventParser(EventParser):
    def wrap_after_month(self, term_name: str) -> int:
        if term_name.startswith("Winter"):
            return 8
        else:
            return 13

    def acad_year_for_term(self, term_name: str) -> int:
        m = re.match(r"\w+ (\d{4})", term_name)
        if m:
            return int(m.group(1))
        return -1

    def events(self, term_name: str) -> Iterable[Event]:
        acad_year = self.acad_year_for_term(term_name)
        html = self.fetcher.fetch_data(term_name)
        soup = BeautifulSoup(html, features="html.parser")
        event_container = cast(Tag, soup.find(class_="views-content"))
        date_nodes = event_container.find_all("h3")
        for date_node in date_nodes:
            description_node = date_node.find_next_sibling("div").find("p")
            date_ranges = parse_html_dates(
                str(date_node.string), acad_year, self.wrap_after_month(term_name)
            )
            for date_range in date_ranges:
                yield Event(description_node.string, date_range[0], date_range[1])
